import logging
from typing import List
from typing import Mapping
from typing import Optional

import service_configuration_lib

from paasta_tools.kubernetes_tools import sanitise_kubernetes_name
from paasta_tools.kubernetes_tools import sanitised_cr_name
from paasta_tools.long_running_service_tools import LongRunningServiceConfig
from paasta_tools.long_running_service_tools import LongRunningServiceConfigDict
from paasta_tools.utils import BranchDictV2
from paasta_tools.utils import compose_job_id
from paasta_tools.utils import decompose_job_id
from paasta_tools.utils import deep_merge_dictionaries
from paasta_tools.utils import DEFAULT_SOA_DIR
from paasta_tools.utils import InvalidJobNameError
from paasta_tools.utils import load_service_instance_config
from paasta_tools.utils import load_v2_deployments_json

KUBERNETES_NAMESPACE = "paasta-vitess"

log = logging.getLogger(__name__)
log.addHandler(logging.NullHandler())


class VitessDeploymentConfigDict(LongRunningServiceConfigDict, total=False):
    replicas: int


class VitessDeploymentConfig(LongRunningServiceConfig):
    config_dict: VitessDeploymentConfigDict

    config_filename_prefix = "vitess"

    def __init__(
        self,
        service: str,
        cluster: str,
        instance: str,
        config_dict: VitessDeploymentConfigDict,
        branch_dict: Optional[BranchDictV2],
        soa_dir: str = DEFAULT_SOA_DIR,
    ) -> None:

        super().__init__(
            cluster=cluster,  # superregion
            instance=instance,  # host-1
            service=service,  # vitess
            soa_dir=soa_dir,
            config_dict=config_dict,
            branch_dict=branch_dict,
        )

    def get_service_name_smartstack(self) -> str:
        """
        We register in vitess.main
        """
        return "vitess_" + self.get_instance()

    def get_nerve_namespace(self) -> str:
        """
        We register in vitess.main
        """
        return "main"

    def get_registrations(self) -> List[str]:
        """
        We register in vitess.main
        """
        registrations = self.config_dict.get("registrations", [])
        for registration in registrations:
            try:
                decompose_job_id(registration)
            except InvalidJobNameError:
                log.error(
                    "Provided registration {} for service "
                    "{} is invalid".format(registration, self.service)
                )

        return registrations or [
            compose_job_id(self.get_service_name_smartstack(), "main")
        ]

    def get_kubernetes_namespace(self) -> str:
        return KUBERNETES_NAMESPACE

    def get_namespace(self) -> str:
        """Get namespace from config, default to 'paasta'"""
        return self.config_dict.get("namespace", KUBERNETES_NAMESPACE)

    def get_instances(self, with_limit: bool = True) -> int:
        return self.config_dict.get("replicas", 1)

    def get_bounce_method(self) -> str:
        """
        Need to map to a paasta bounce method and crossover is the closest
        """
        return "crossover"

    def get_sanitised_service_name(self) -> str:
        return sanitise_kubernetes_name(self.get_service())

    def get_sanitised_instance_name(self) -> str:
        return sanitise_kubernetes_name(self.get_instance())

    def get_sanitised_deployment_name(self) -> str:
        return self.get_sanitised_instance_name()

    def validate(
        self,
        params: List[str] = [
            "cpus",
            "security",
            "dependencies_reference",
            "deploy_group",
        ],
    ) -> List[str]:
        # Use InstanceConfig to validate shared config keys like cpus and mem
        # TODO: add mem back to this list once we fix PAASTA-15582 and
        # move to using the same units as flink/marathon etc.
        error_msgs = super().validate(params=params)

        if error_msgs:
            name = self.get_instance()
            return [f"{name}: {msg}" for msg in error_msgs]
        else:
            return []


def load_vitess_instance_config(
    service: str,
    instance: str,
    cluster: str,
    load_deployments: bool = True,
    soa_dir: str = DEFAULT_SOA_DIR,
) -> VitessDeploymentConfig:
    general_config = service_configuration_lib.read_service_configuration(
        service, soa_dir=soa_dir
    )
    instance_config = load_service_instance_config(
        service, instance, " vitesscluster", cluster, soa_dir=soa_dir
    )
    general_config = deep_merge_dictionaries(
        overrides=instance_config, defaults=general_config
    )

    branch_dict: Optional[BranchDictV2] = None
    if load_deployments:
        deployments_json = load_v2_deployments_json(service, soa_dir=soa_dir)
        temp_instance_config = VitessDeploymentConfig(
            service=service,
            cluster=cluster,
            instance=instance,
            config_dict=general_config,
            branch_dict=None,
            soa_dir=soa_dir,
        )
        branch = temp_instance_config.get_branch()
        deploy_group = temp_instance_config.get_deploy_group()
        branch_dict = deployments_json.get_branch_dict(service, branch, deploy_group)

    return VitessDeploymentConfig(
        service=service,
        cluster=cluster,
        instance=instance,
        config_dict=general_config,
        branch_dict=branch_dict,
        soa_dir=soa_dir,
    )


# TODO: read this from CRD in service configs
def cr_id(service: str, instance: str) -> Mapping[str, str]:
    return dict(
        group="yelp.com",
        version="v1alpha1",
        namespace=KUBERNETES_NAMESPACE,
        plural="vitess",
        name=sanitised_cr_name(service, instance),
    )



# Copyright 2015-2016 Yelp Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import re
from typing import Dict

import aiohttp
from mypy_extensions import TypedDict

from paasta_tools.utils import get_user_agent

HACHECK_TIMEOUT = aiohttp.ClientTimeout(total=45, connect=30, sock_read=10)

SpoolInfo = TypedDict(
    "SpoolInfo",
    {"service": str, "state": str, "since": float, "until": float, "reason": str},
    total=False,
)


async def post_spool(url: str, status: str, data: Dict["str", "str"]) -> None:
    async with aiohttp.ClientSession(timeout=HACHECK_TIMEOUT) as session:
        async with session.post(
            url, data=data, headers={"User-Agent": get_user_agent()}
        ) as resp:
            resp.raise_for_status()


async def get_spool(spool_url: str) -> SpoolInfo:
    """Query hacheck for the state of a task, and parse the result into a dictionary."""
    if spool_url is None:
        return None

    # TODO: aiohttp says not to create a session per request. Fix this.
    async with aiohttp.ClientSession(timeout=HACHECK_TIMEOUT) as session:
        async with session.get(
            spool_url, headers={"User-Agent": get_user_agent()}
        ) as response:
            if response.status == 200:
                return {"state": "up"}

            regex = "".join(
                [
                    "^",
                    r"Service (?P<service>.+)",
                    r" in (?P<state>.+) state",
                    r"(?: since (?P<since>[0-9.]+))?",
                    r"(?: until (?P<until>[0-9.]+))?",
                    r"(?:: (?P<reason>.*))?",
                    "$",
                ]
            )

            response_text = await response.text()
            match = re.match(regex, response_text)
            groupdict = match.groupdict()
            info: SpoolInfo = {}
            info["service"] = groupdict["service"]
            info["state"] = groupdict["state"]
            if "since" in groupdict:
                info["since"] = float(groupdict["since"] or 0)
            if "until" in groupdict:
                info["until"] = float(groupdict["until"] or 0)
            if "reason" in groupdict:
                info["reason"] = groupdict["reason"]
            return info



#!/usr/bin/env python3.8
import itertools
import json
import sys
from collections import defaultdict
from datetime import datetime


def get_datetime_from_ts(ts):
    tformat = "%Y-%m-%dT%H:%M:%S.%f"
    return datetime.strptime(ts, tformat)


def get_deploy_durations_from_file(filename):
    """
    filename: path to a file to be parsed for datetime data
    The expected input is a paasta service log for the deploy events
    The way I've been fetching them is by running 'internal logreader command' | grep deploy | grep event > filename
    """
    file_object = open(filename, "r")
    data = sorted(
        [json.loads(line.rstrip("\n")) for line in file_object],
        key=lambda x: get_datetime_from_ts(x["timestamp"]),
    )

    timedeltas = defaultdict(list)
    last_time = dict()
    instance_bitvector = defaultdict(bool)  # defaults to False

    for datum in data:
        time = get_datetime_from_ts(datum["timestamp"])
        instance = datum["instance"]
        if "in progress" in datum["message"] and not instance_bitvector[instance]:
            instance_bitvector[instance] = True
            last_time[instance] = time
        elif "finishing" in datum["message"]:
            instance_bitvector[instance] = False
            timedeltas[instance].append(time - last_time[instance])

    return timedeltas


def display_bounce_info(timedeltas):
    """
    timedeltas: iterable of timedelta objects
    """
    std = list(sorted(timedeltas))
    print("Median time to bounce: {} seconds".format(std[len(std) / 2]))
    print("10% time to bounce: {}".format(std[len(std) / 10]))
    print("90% time to bounce: {}".format(std[len(std) * 9 / 10]))


def main(filenames):
    for filename in filenames:
        print(filename)
        print("=========================")
        timedeltas = get_deploy_durations_from_file(filename)
        for instance, tdlist in timedeltas.items():
            if timedeltas:
                print("Instance: %s" % instance)
                display_bounce_info(tdlist)
        print("Overall:")
        display_bounce_info(itertools.chain.from_iterable(timedeltas.values()))
        print("=========================")


if __name__ == "__main__":
    main(filenames=sys.argv[1:])



#!/usr/bin/env python
# Copyright 2015-2016 Yelp Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import argparse
import datetime
import logging

import dateutil.parser
from dateutil import tz
from pytimeparse import timeparse

from paasta_tools import marathon_tools


log = logging.getLogger(__name__)


def parse_args():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "-a",
        "--age",
        dest="age",
        type=timedelta_type,
        default="1h",
        help="Max age of a Marathon deployment before it is stopped."
        "Any pytimeparse unit is supported",
    )
    parser.add_argument(
        "-n",
        "--dry-run",
        action="store_true",
        help="Don't actually stop any Marathon deployments",
    )
    parser.add_argument("-v", "--verbose", action="store_true")
    options = parser.parse_args()
    return options


def timedelta_type(value):
    """Return the :class:`datetime.datetime.DateTime` for a time in the past.
    :param value: a string containing a time format supported by :mod:`pytimeparse`
    """
    if value is None:
        return None
    return datetime_seconds_ago(timeparse.timeparse(value))


def datetime_seconds_ago(seconds):
    return now() - datetime.timedelta(seconds=seconds)


def now():
    return datetime.datetime.now(tz.tzutc())


def delete_deployment_if_too_old(client, deployment, max_date, dry_run):
    started_at = dateutil.parser.parse(deployment.version)
    age = now() - started_at
    if started_at < max_date:
        if dry_run is True:
            log.warning(
                f"Would delete {deployment.id} for {deployment.affected_apps[0]} as it is {age} old"
            )
        else:
            log.warning(
                f"Deleting {deployment.id} for {deployment.affected_apps[0]} as it is {age} old"
            )
            client.delete_deployment(deployment_id=deployment.id, force=False)
    else:
        if dry_run is True:
            log.warning(
                f"NOT deleting {deployment.id} for {deployment.affected_apps[0]} as it is {age} old"
            )


def main():
    args = parse_args()
    if args.verbose:
        logging.basicConfig(level=logging.DEBUG)
    else:
        logging.basicConfig(level=logging.WARNING)

    clients = marathon_tools.get_list_of_marathon_clients()

    for client in clients:
        for deployment in client.list_deployments():
            delete_deployment_if_too_old(
                client=client,
                deployment=deployment,
                max_date=args.age,
                dry_run=args.dry_run,
            )


if __name__ == "__main__":
    main()
